from collections import defaultdict
import logging
from typing import List, Optional, Dict
import time

import win32pdh

default_logger = logging.getLogger(__name__)


class Counter:
    # TODO: This needs to be split into Metric, Instance, Collector
    # TODO: Collector has the key and aggregation
    # TODO: Metric has the Object and Counter, and List of Instance
    def __init__(
        self,
        obj: str,
        counter: str,
        instance: Optional[str],
        handle,
        key: str,
        metric_type: str,
        aggregation: str,
        dimension_name: str = None,
    ):
        self.path = f"{obj}({instance})/{counter}" if instance is not None else f"{obj}/{counter}"
        self.obj = obj
        self.instance = instance
        self.counter = counter
        self.handle = handle
        self.key = key
        self.metric_type = metric_type
        self.value = 0
        self.aggregation = aggregation
        self.dimension_name = dimension_name

    def __str__(self):
        return f"Counter({self.path}, {self.handle}, {self.key}, value: {self.value})"

    def __repr__(self):
        return self.__str__()


class Perfmon:
    def __init__(self, logger: logging.Logger = default_logger):
        self.logger = logger
        self.counters: List[Counter] = []

    def __enter__(self):
        self.query = win32pdh.OpenQuery()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.query is not None:
            win32pdh.CloseQuery(self.query)

    def get_counters_per_key(self) -> Dict[str, List[Counter]]:
        metrics_dict: Dict[str, List[Counter]] = defaultdict(list)

        for counter in self.counters:
            metrics_dict[counter.key].append(counter)

        return metrics_dict

    def _has_instances(self, obj) -> bool:
        try:
            _, instances = win32pdh.EnumObjectItems(None, None, obj, win32pdh.PERF_DETAIL_WIZARD)
            if instances:
                return True
        except Exception as e:
            return False

        return False

    def _add_counter_path(
        self, obj: str, counter: str, instance: Optional[str], key: str, aggregation, metric_type="absolute", dimension_name=None
    ):
        if instance is None and aggregation == "count":
            self.counters.append(
                Counter(obj, counter, instance, None, key, metric_type, aggregation, dimension_name=dimension_name)
            )
        else:
            path = win32pdh.MakeCounterPath((None, obj, instance, None, -1, counter))
            counter_handle = win32pdh.AddEnglishCounter(self.query, path)
            localized_path_elements = win32pdh.GetCounterInfo(counter_handle)
            localized_path = localized_path_elements[6]
            expanded_paths = win32pdh.ExpandCounterPath(localized_path)
            for path in expanded_paths:
                if self.logger.level == logging.DEBUG:
                    self.logger.info(f"Adding localized path: {path}")
                counter_handle = win32pdh.AddCounter(self.query, path)
                counter_elements = win32pdh.GetCounterInfo(counter_handle)
                instance_data = counter_elements[7]
                instance_name = "Unknown"
                if instance_data:
                    instance_name = instance_data[2]

                self.counters.append(
                    Counter(obj, counter, instance_name, counter_handle, key, metric_type, aggregation,
                            dimension_name=dimension_name)
                )

    def add_counter(self, obj: str, counter: str, key="", aggregation="avg", metric_type="absolute", dimension_name=None):
        try:
            has_instances = self._has_instances(obj)
            instances = "*" if has_instances else None
            if self.logger.level == logging.DEBUG:
                self.logger.info(f"Attempting to add counter, object='{obj}', counter='{counter}, instances="
                                 f"'{instances}'")
            self._add_counter_path(obj, counter, instances, key, aggregation, metric_type, dimension_name=dimension_name)

        except Exception as e:
            self.logger.warning(f"Could not add counter: 'obj={obj},counter={counter}': {e}")

    def collect_data(self):
        try:
            win32pdh.CollectQueryData(self.query)
            time.sleep(1)
            win32pdh.CollectQueryData(self.query)

            for counter in self.counters:
                try:
                    if counter.handle is not None:
                        counter.value = win32pdh.GetFormattedCounterValue(counter.handle, win32pdh.PDH_FMT_DOUBLE)[1]
                except Exception as e:
                    self.logger.error(f"Could not obtain value for counter {counter}: {e}")
        except Exception as e:
            self.logger.error(f"Could not call CollectQueryData: {e}")


def main():
    logger = logging.getLogger(__name__)
    st = logging.StreamHandler()
    logger.setLevel(logging.DEBUG)
    fmt = logging.Formatter("%(asctime)s - %(levelname)s - %(funcName)s - %(message)s ")
    st.setFormatter(fmt)
    logger.addHandler(st)

    with Perfmon(logger) as p:
        p.add_counter("LogicalDisk", "% Free Space", "disk_free")
        p.add_counter("Memory", "Available Bytes", "memory")
        p.add_counter("Process", "Handle Count", "process")
        p.collect_data()

        for metric, counters in p.get_counters_per_key().items():
            for counter in counters:
                print(f"{metric} - {counter.counter} - {counter.instance} = {counter.value}")


if __name__ == "__main__":
    main()
